import os
import sys

from pip._internal.operations.freeze import freeze as pip_freeze


def lines_from_file(filename):
    with open(filename) as f:
        return f.read().split("\n")


def read_pip(filename):
    """Return lines in pip file, concatenating included requirement files."""
    lines = lines_from_file(filename)
    for line in lines:
        if line.startswith("-r "):
            orig_dirpath = os.path.dirname(os.path.realpath(filename))
            sub_filename = line.split(" ")[1]
            sub_filepath = os.path.join(orig_dirpath, sub_filename)
            lines.extend(read_pip(sub_filepath))
    return lines


def get_package_versions(lines, ignore_external_and_at=False):
    """Return a dictionary of package versions."""
    versions = {}
    for line in lines:
        line = line.strip()

        if len(line) == 0 or line.startswith("#") or line.startswith("-r "):
            continue

        if line.startswith("https://"):
            continue

        if ignore_external_and_at:
            if line.startswith("-e") or " @ " in line:
                continue

        full_name, version_and_extras = line.split("==", 1)
        # Strip extras
        name = full_name.split("[", 1)[0].lower()
        version = version_and_extras.split(" ", 1)[0]
        versions[name] = version

    return versions


def get_mismatches(requirements_file_path):
    """Return a dictionary of requirement mismatches."""
    pip_lines = read_pip(requirements_file_path)

    expected = get_package_versions(pip_lines)
    actual = get_package_versions(pip_freeze(), ignore_external_and_at=True)

    mismatches = {}
    for name, version in expected.items():
        if name not in actual:
            mismatches[name] = (version, None)
            continue

        if version != actual[name]:
            mismatches[name] = (version, actual[name])

    return mismatches


def print_errors(errors, pre_text=None, post_text=None):
    """Print list of errors to stderr with optional pre_text and post_text."""
    sys.stderr.write("\033[91m")  # red text
    if pre_text:
        sys.stderr.write(pre_text + "\n")
    for message in errors:
        sys.stderr.write("    * {}\n".format(message))
    if post_text:
        sys.stderr.write(post_text)
    sys.stderr.write("\033[0m")


def check_requirements(requirements_file_path, post_text=None):
    """
    Print errors and exit program if there are mismatches with the requirements
    file.
    """
    mismatches = get_mismatches(requirements_file_path)
    if mismatches:
        errors = []
        for name, (expected, actual) in mismatches.items():
            if actual is None:
                errors.append(
                    "Package {} is in {} but not in virtualenv".format(
                        name, requirements_file_path
                    )
                )
                continue

            if expected != actual:
                msg = (
                    "Package {0} has version {1} but you have version {2} "
                    + "installed."
                ).format(name, expected, actual)
                errors.append(msg)

        print_errors(
            errors,
            "There are requirement mismatches with {}".format(requirements_file_path),
            post_text,
        )
        sys.exit(1)



from unittest import mock

import pytest

from pip_lock import (
    check_requirements,
    get_mismatches,
    get_package_versions,
    print_errors,
    read_pip,
)


def create_file(tmpdir, name, text):
    t = tmpdir.join(name)
    t.write(text)
    return str(t)


class TestReadPip:
    def test_read(self, tmpdir):
        path = create_file(tmpdir, "requirements.txt", "package1==1.0\npackage2==1.1")
        assert read_pip(path) == ["package1==1.0", "package2==1.1"]

    def test_include(self, tmpdir):
        inc_path = create_file(tmpdir, "requirements_inc.txt", "other-package==1.0")
        path = create_file(tmpdir, "requirements.txt", "-r {}".format(inc_path))

        assert read_pip(path) == ["-r {}".format(inc_path), "other-package==1.0"]

    def test_empty(self, tmpdir):
        path = create_file(tmpdir, "requirements.txt", "")
        assert read_pip(path) == [""]


class TestGetPackageVersion:
    def test_version(self):
        assert get_package_versions(["package1==1.0", "package2==1.1"]) == {
            "package1": "1.0",
            "package2": "1.1",
        }

    def test_ignore_empty(self):
        assert get_package_versions([""]) == {}

    def test_ignore_comments(self):
        assert get_package_versions(["# Comment"]) == {}

    def test_ignore_includes(self):
        assert get_package_versions(["-r example.txt"]) == {}

    def test_ignore_urls(self):
        assert get_package_versions(["https://www.google.com"]) == {}


class TestGetMismatches:
    def setUp(self, tmpdir):
        super().setUp()
        self.requirements_path = create_file(
            tmpdir, "requirements.txt", "package1==1.1\npackage2==1.2"
        )

    @mock.patch("pip_lock.pip_freeze")
    def test_relative_requirements_file(self, pip_freeze, tmpdir):
        pip_freeze.return_value = ["package==1.1"]
        create_file(tmpdir, "requirements.txt", "package==1.2")
        with tmpdir.as_cwd():
            assert get_mismatches("requirements.txt") == {"package": ("1.2", "1.1")}

    @mock.patch("pip_lock.pip_freeze")
    def test_version_mismatch(self, pip_freeze, tmpdir):
        pip_freeze.return_value = ["package==1.1"]
        requirements_path = create_file(tmpdir, "requirements.txt", "package==1.2")

        assert get_mismatches(requirements_path) == {"package": ("1.2", "1.1")}

    @mock.patch("pip_lock.pip_freeze")
    def test_missing(self, pip_freeze, tmpdir):
        pip_freeze.return_value = [""]
        requirements_path = create_file(tmpdir, "requirements.txt", "package==1.1")

        assert get_mismatches(requirements_path) == {"package": ("1.1", None)}

    @mock.patch("pip_lock.pip_freeze")
    def test_no_mismatches(self, pip_freeze, tmpdir):
        pip_freeze.return_value = ["package==1.1"]
        requirements_path = create_file(tmpdir, "requirements.txt", "package==1.1")

        assert get_mismatches(requirements_path) == {}

    @mock.patch("pip_lock.pip_freeze")
    def test_no_mismatches_case_insensitive(self, pip_freeze, tmpdir):
        pip_freeze.return_value = ["Package==1.1"]
        requirements_path = create_file(tmpdir, "requirements.txt", "package==1.1")

        assert get_mismatches(requirements_path) == {}

    @mock.patch("pip_lock.pip_freeze")
    def test_empty(self, pip_freeze, tmpdir):
        pip_freeze.return_value = [""]
        requirements_path = create_file(tmpdir, "requirements.txt", "")

        assert get_mismatches(requirements_path) == {}

    @mock.patch("pip_lock.pip_freeze")
    def test_editable_packages_ignored(self, pip_freeze, tmpdir):
        pip_freeze.return_value = [
            (
                "-e git+git@github.com:adamchainz/pip-lock.git@"
                + "efac0eef8072d73b001b1bae0731c1d58790ac4b#egg=pip-lock"
            ),
            "package==1.1",
        ]
        requirements_path = create_file(tmpdir, "requirements.txt", "package==1.1")

        assert get_mismatches(requirements_path) == {}

    @mock.patch("pip_lock.pip_freeze")
    def test_at_packages_ignored(self, pip_freeze, tmpdir):
        pip_freeze.return_value = [
            "pip @ file:///tmp/pip-20.1.1-py2.py3-none-any.whl",
            "package==1.1",
        ]
        requirements_path = create_file(tmpdir, "requirements.txt", "package==1.1")

        assert get_mismatches(requirements_path) == {}

    @mock.patch("pip_lock.pip_freeze")
    def test_package_with_extra(self, pip_freeze, tmpdir):
        pip_freeze.return_value = ["package==1.1"]
        requirements_path = create_file(
            tmpdir, "requirements.txt", "package[anextra]==1.1"
        )

        assert get_mismatches(requirements_path) == {}


class TestPrintErrors:
    def test_errors(self, capsys):
        print_errors(["error message 1", "error message 2"])
        _, err = capsys.readouterr()
        assert "error message 1" in err
        assert "error message 2" in err

    def test_pre_post_text(self, capsys):
        print_errors(["error message"], "pre text", "post text")
        _, err = capsys.readouterr()
        assert "error message" in err
        assert "pre text" in err
        assert "post text" in err


class TestCheckRequirements:
    @mock.patch("pip_lock.get_mismatches")
    def test_no_mismatches(self, get_mismatches):
        get_mismatches.return_value = {}
        check_requirements("requirements.txt")

    @mock.patch("pip_lock.get_mismatches")
    def test_mismatches(self, get_mismatches, capsys):
        get_mismatches.return_value = {
            "package1": ("1.1", "1.0"),
            "package2": ("1.0", None),
        }
        with pytest.raises(SystemExit):
            check_requirements("requirements.txt")

        _, err = capsys.readouterr()
        assert "package1 has version 1.1 but you have version 1.0 installed" in err
        assert "package2 is in requirements.txt but not in virtualenv" in err

    @mock.patch("pip_lock.pip_freeze")
    def test_relative_requirements_file(self, pip_freeze, tmpdir):
        pip_freeze.return_value = ["package==1.2"]
        create_file(tmpdir, "requirements.txt", "package==1.2")
        with tmpdir.as_cwd():
            check_requirements("requirements.txt")
